diff --exclude=CVS -ur u-boot-cvs/drivers/cfi_flash.c u-boot/drivers/cfi_flash.c
--- u-boot-cvs/drivers/cfi_flash.c	2004-02-11 19:47:14.000000000 -0500
+++ u-boot/drivers/cfi_flash.c	2004-02-21 00:00:56.000000000 -0500
@@ -45,7 +45,10 @@
 #undef	DEBUG
 #include <common.h>
 #include <asm/processor.h>
+#include <linux/byteorder/swab.h>
+#include <malloc.h>
 #ifdef	CFG_FLASH_CFI_DRIVER
+
 /*
  * This file implements a Common Flash Interface (CFI) driver for U-Boot.
  * The width of the port and the width of the chips are determined at initialization.
@@ -161,11 +164,14 @@
 
 typedef unsigned long flash_sect_t;
 
-static void flash_add_byte (flash_info_t * info, cfiword_t * cword, uchar c);
+static void flash_unlock_seq (flash_info_t * info, flash_sect_t sect);
 static void flash_make_cmd (flash_info_t * info, uchar cmd, void *cmdbuf);
 static void flash_write_cmd (flash_info_t * info, flash_sect_t sect,
 			     uint offset, uchar cmd);
-static void flash_unlock_seq (flash_info_t * info, flash_sect_t sect);
+
+static ulong flash_get_size (ulong base, int banknum);
+
+static int flash_write_aligned(flash_info_t * info, ulong addr, uchar * data, ulong words);
 static int flash_isequal (flash_info_t * info, flash_sect_t sect, uint offset,
 			  uchar cmd);
 static int flash_isset (flash_info_t * info, flash_sect_t sect, uint offset,
@@ -173,7 +179,6 @@
 static int flash_toggle (flash_info_t * info, flash_sect_t sect, uint offset,
 			 uchar cmd);
 static int flash_detect_cfi (flash_info_t * info);
-static ulong flash_get_size (ulong base, int banknum);
 static int flash_write_cfiword (flash_info_t * info, ulong dest,
 				cfiword_t cword);
 static int flash_full_status_check (flash_info_t * info, flash_sect_t sector,
@@ -475,103 +480,209 @@
  * 1 - write timeout
  * 2 - Flash not erased
  */
+ 
+#define READ_CURR(x,addr)	({						\
+	x = -1;									\
+	switch (info->portwidth) {						\
+		case FLASH_CFI_8BIT:						\
+			x = (unsigned long long)(*((uchar*)addr));		\
+			break;							\
+		case FLASH_CFI_16BIT:						\
+			x = (unsigned long long)(*((unsigned short*)addr));	\
+			break;							\
+		case FLASH_CFI_32BIT:						\
+			x = (unsigned long long)(*((unsigned long*)addr));	\
+			break;							\
+		case FLASH_CFI_64BIT:						\
+			x = (unsigned long long)(*((unsigned long long*)addr));	\
+			break;							\
+	}									\
+	})
+			
 int write_buff (flash_info_t * info, uchar * src, ulong addr, ulong cnt)
 {
-	ulong wp;
-	ulong cp;
-	int aln;
-	cfiword_t cword;
+	unsigned long tmp, mask, curr;
+	uchar * buff1, * buff2, * t;
 	int i, rc;
-
-#ifdef CFG_FLASH_USE_BUFFER_WRITE
-	int buffered_size;
-#endif
-	int x8mode = 0;
-
-	/* special handling of 16 bit devices in 8 bit mode */
-	if ((info->interface == FLASH_CFI_X8X16)
-	    && (info->chipwidth == FLASH_CFI_BY8)) {
-		switch (info->vendor) {
-		case CFI_CMDSET_INTEL_STANDARD:
-		case CFI_CMDSET_INTEL_EXTENDED:
-			x8mode = info->portwidth;
-			info->portwidth >>= 1;	/* XXX - Need to test on x9/x16 in parallel. */
-			/*info->portwidth = FLASH_CFI_8BIT; */ /* XXX - Need to test on x9/x16 in parallel. */
+	int divisor;
+		
+	switch (info->portwidth) {
+		case FLASH_CFI_8BIT:
+			divisor = 1;
 			break;
-		case CFI_CMDSET_AMD_STANDARD:
-		case CFI_CMDSET_AMD_EXTENDED:
-		default:
+		case FLASH_CFI_16BIT:
+			divisor = 2;
 			break;
-		}
+		case FLASH_CFI_32BIT:
+			divisor = 4;
+			break;
+		case FLASH_CFI_64BIT:
+			divisor = 8;
+			break;
+		default:
+			printf("cfi_flash:write_buff:UNKNOWN PORT WIDTH, FAILING\n");
+			return 1;
 	}
-	/* get lower aligned address */
-	/* get lower aligned address */
-	wp = (addr & ~(info->portwidth - 1));
 
-	/* handle unaligned start */
-	if ((aln = addr - wp) != 0) {
-		cword.l = 0;
-		cp = wp;
-		for (i = 0; i < aln; ++i, ++cp)
-			flash_add_byte (info, &cword, (*(uchar *) cp));
-
-		for (; (i < info->portwidth) && (cnt > 0); i++) {
-			flash_add_byte (info, &cword, *src++);
-			cnt--;
-			cp++;
-		}
-		for (; (cnt == 0) && (i < info->portwidth); ++i, ++cp)
-			flash_add_byte (info, &cword, (*(uchar *) cp));
-		if ((rc = flash_write_cfiword (info, wp, cword)) != 0)
-			return rc;
-		wp = cp;
-	}
+#ifdef DEBUG
+	printf("cfi_flash: write_buff: info 0x%p src 0x%p addr 0x%08lx cnt 0x%08lx, divisor = %d\n",
+		info, src, addr, cnt, divisor );
+#endif
 
-	/* handle the aligned part */
-#ifdef CFG_FLASH_USE_BUFFER_WRITE
-	buffered_size = (info->portwidth / info->chipwidth);
-	buffered_size *= info->buffer_size;
-	while (cnt >= info->portwidth) {
-		i = buffered_size > cnt ? cnt : buffered_size;
-		if ((rc = flash_write_cfibuffer (info, wp, src, i)) != ERR_OK)
-			return rc;
-		wp += i;
-		src += i;
-		cnt -= i;
+	/*
+	 * head and tail "padding" as previously implemented is BAD.  it leads
+	 * to broken images.  handle this in a smarter way. 
+	 */
+	 
+	/* verify all of the target flash is erased */ 
+	for (t = (uchar*)addr; t < (uchar*)(addr+cnt); t++)
+	{
+		if (*t != 0xFF)
+		{
+			printf("FAILURE: flash at 0x%p not erased!\n", t);
+			return 2;
+		}
 	}
-#else
-	while (cnt >= info->portwidth) {
-		cword.l = 0;
-		for (i = 0; i < info->portwidth; i++) {
-			flash_add_byte (info, &cword, *src++);
+		
+	/* 
+	 * deal with "leading" unalignment in the _FLASH_ target -- ie, we 
+	 * need to be able to write to _ANY_ address, not just 2- or 4-byte
+	 * aligned ones... we want to put ourselves on an N-byte aligned 
+	 * spot to do rapid flash filling, however.  so do initial write.
+	 */
+	
+	i = addr & (divisor-1);
+	if (i)
+	{
+		rc = i;
+		addr &= ~(divisor-1);			/* find the base tgt address	*/
+		tmp = 0;
+		mask = -1;				/* all 1's			*/
+		for (; (i > 0) && (cnt > 0); i--, cnt--)
+		{
+			tmp  <<= 8;
+			mask <<= 8;
+			tmp |= src[0];			/* load tmp with low-bits to write */
+			src++;
 		}
-		if ((rc = flash_write_cfiword (info, wp, cword)) != 0)
+		READ_CURR(curr,addr);
+#ifdef DEBUG
+		printf("WB: misalign head: cnt %d curr %08lx addr %08lx tmp %08lx mask %08lx\n",
+			rc, curr, addr, tmp, mask);
+#endif
+		curr = (curr & mask) | tmp;		/* mask in new low-bits 	*/
+#ifdef __LITTLE_ENDIAN
+		curr = __swab32( curr );
+#endif	
+		if ((rc = flash_write_aligned(info, addr, (uchar*)&curr, divisor)) 
+				!= ERR_OK)
+		{
+			printf("cfi_flash_write: leading Flash misalignment fix failed\n");
 			return rc;
-		wp += info->portwidth;
-		cnt -= info->portwidth;
-	}
-#endif /* CFG_FLASH_USE_BUFFER_WRITE */
-	if (cnt == 0) {
-		return (0);
+		}
+		if (cnt == 0)				/* did we finish? 		*/
+			return 0;
+		addr += divisor;			/* okay, move to aligned block	*/
 	}
+	
+#ifdef DEBUG
+	printf("WB: A: info 0x%p src 0x%p addr 0x%08lx cnt 0x%08lx, divisor = %d\n",
+		info, src, addr, cnt, divisor );
+#endif
 
-	/*
-	 * handle unaligned tail bytes
+	/* 
+	 * okay, we need to get our "source" buffer of data to be 4-byte aligned to
+	 * reduce headaches and other subtle problems.  do a little hackery.
 	 */
-	cword.l = 0;
-	for (i = 0, cp = wp; (i < info->portwidth) && (cnt > 0); ++i, ++cp) {
-		flash_add_byte (info, &cword, *src++);
-		--cnt;
-	}
-	for (; i < info->portwidth; ++i, ++cp) {
-		flash_add_byte (info, &cword, (*(uchar *) cp));
+	if (((ulong)src) & (divisor-1))
+	{
+		buff1 = (uchar *)malloc( cnt + 20 );	/* what's left + margins	*/
+		if (buff1 == NULL)
+		{
+			printf("cfi_flash:write_buff:FAILED TO MALLOC ALIGNED BUFFER\n");
+			return 1;
+		}
+		buff2 = (uchar *)( (((ulong)buff1) + (divisor-1)) & ~(divisor-1) );
+		for (i=0; i<cnt; i++)
+			buff2[i] = src[i];
 	}
+	else
+	{
+		buff2 = src;
+		buff1 = NULL;
+	}
+	
+	/* now, calculate and pump into flash our (flash + src) aligned addr/data 	*/
+	
+#ifdef DEBUG
+	printf("WB: B: info 0x%p src 0x%p srcAL 0x%p addr 0x%08lx cnt 0x%08lx, divisor = %d\n",
+		info, src, buff2, addr, cnt, divisor );
+#endif
+
+	i = (cnt / divisor);
+	if ((rc = flash_write_aligned( info, addr, buff2, i*divisor )))
+	{
+		printf("cfi_flash: write_buff: failed to write ALIGNED region!\n");
+		if (buff1 != NULL)
+			free( buff1 );
+		return rc;
+	}
+	
+	/* any trailing data to muck with? */
+	
+	addr  += i * divisor;
+	buff2 += i * divisor;
+	cnt   -= i * divisor;
+
+#ifdef DEBUG
+	printf("WB: C: info 0x%p src 0x%p addr 0x%08lx cnt 0x%08lx, divisor = %d\n",
+		info, buff2, addr, cnt, divisor );
+#endif
 
-	/* special handling of 16 bit devices in 8 bit mode */
-	if (x8mode) {
-		info->portwidth = x8mode;;
+	if (cnt == 0)					/* are we done?			*/
+	{
+		if (buff1 != NULL)
+			free( buff1 );
+		return 0;				/* we're done! yay!		*/
+	}
+	
+	/* well, we know the remaining "cnt" is < divisor, ie, will fit in one 		*/
+	/* "divisor" variable... so hack up a fix like the beginning value fix		*/
+
+	tmp = 0;
+	mask = ((divisor-1)*8);
+	for (i = cnt; i > 0; i--)
+	{
+		tmp |= (((ulong)buff2[0]) << mask);	/* shift up to proper spot	*/
+		mask -= 8;				/* drop by 8			*/
+		buff2++;
+	}
+	
+	mask = -1;					/* all 1's 			*/
+	for (i = cnt; i > 0; i--)
+		mask >>= 8;
+		
+	READ_CURR(curr,addr);				/* get what's already there 	*/
+#ifdef DEBUG
+	printf("WB: misalign tail: cnt %08lx curr %08lx addr %08lx tmp %08lx mask %08lx\n",
+		cnt, curr, addr, tmp, mask);
+#endif
+	curr = (curr & mask) | tmp;			/* mask in new low-bits 	*/
+#ifdef __LITTLE_ENDIAN
+	curr = __swab32( curr );
+#endif	
+	if ((rc = flash_write_aligned(info, addr, (uchar*)&curr, divisor)) 
+			!= ERR_OK)
+	{
+		printf("cfi_flash_write: trailing Flash misalignment fix failed\n");
+		if (buff1 != NULL)
+			free( buff1 );
+		return rc;
 	}
-	return flash_write_cfiword (info, wp, cword);
+
+	if (buff1 != NULL)
+		free( buff1 );	
+	return 0;
 }
 
 /*-----------------------------------------------------------------------
@@ -753,6 +864,44 @@
 }
 
 
+static int flash_write_aligned(flash_info_t * info, ulong addr, uchar * data, ulong cnt)
+{
+	ulong i, buffered_size;
+	int rc;
+
+#ifdef CFG_FLASH_USE_BUFFER_WRITE
+	buffered_size = (info->portwidth / info->chipwidth);
+	buffered_size *= info->buffer_size;
+	while (cnt >= info->portwidth) {
+		i = buffered_size > cnt ? cnt : buffered_size;
+		if ((rc = flash_write_cfibuffer (info, addr, data, i)) != ERR_OK)
+		{
+			printf("cfi_flash:write_aligned:FAILED in BUFFER WRITE\n");
+			return rc;
+		}
+		addr += i;
+		data += i;
+		cnt -= i;
+	}
+#else
+	cfiword_t cword;
+
+	while (cnt >= info->portwidth) {
+		cword.l = 0;
+		for (i = 0; i < info->portwidth; i++) 
+		{
+			flash_add_byte (info, &cword, *data++);
+		}
+		if ((rc = flash_write_cfiword (info, addr, cword)) != 0)
+			return rc;
+		addr += info->portwidth;
+		cnt -= info->portwidth;
+	}
+#endif
+	return 0;		/* all OK */
+}
+
+
 /*-----------------------------------------------------------------------
  * make a proper sized command based on the port and chip widths
  */
@@ -761,16 +910,29 @@
 	int i;
 
 #if defined(__LITTLE_ENDIAN)
-	ushort stmp;
+	ushort stmpw;
+	uint   stmpi;
 #endif
 	uchar *cp = (uchar *) cmdbuf;
 
 	for (i = 0; i < info->portwidth; i++)
 		*cp++ = ((i + 1) % info->chipwidth) ? '\0' : cmd;
 #if defined(__LITTLE_ENDIAN)
-	if (info->portwidth == 2) {
-		stmp = *(ushort *) cmdbuf;
-		*(ushort *) cmdbuf = swab16 (stmp);
+	switch (info->portwidth)
+	{
+		case FLASH_CFI_8BIT:
+			break;
+		case FLASH_CFI_16BIT:
+			stmpw = *(ushort *) cmdbuf;
+			*(ushort *) cmdbuf = __swab16 (stmpw);
+			break;
+		case FLASH_CFI_32BIT:
+			stmpi = *(uint *) cmdbuf;
+			*(uint *) cmdbuf = __swab32 (stmpi);
+			break;
+		default:
+			printf("WARNING: flash_make_cmd: unsuppport LittleEndian mode\n");
+			break;
 	}
 #endif
 }
